"""
Test suite for jpy-sql-runner SQL helper module.

Comprehensive unit tests for SQL parsing, comment removal, statement splitting,
and statement type detection functionality.

Copyright (c) 2025, Jim Schilling

This module is licensed under the MIT License.
"""

import unittest
from jpy_sql_runner.sql_helper import (
    remove_sql_comments,
    parse_sql_statements,
    split_sql_file,
    detect_statement_type,
)
import sqlparse
from sqlparse.sql import Token, Statement


class TestSqlHelper(unittest.TestCase):
    """Comprehensive tests for SQL helper functions."""

    def test_remove_sql_comments(self):
        """Test SQL comment removal functionality."""
        # Test single-line comments
        sql_with_single_comments = """
        CREATE TABLE users (
            id INTEGER PRIMARY KEY, -- user id
            name TEXT NOT NULL,     -- user name
            email TEXT              -- user email
        );
        """
        clean_sql = remove_sql_comments(sql_with_single_comments)
        self.assertNotIn("--", clean_sql)
        self.assertIn("CREATE TABLE", clean_sql)

        # Test multi-line comments
        sql_with_multi_comments = """
        /* This is a multi-line comment */
        CREATE TABLE products (
            id INTEGER PRIMARY KEY, /* product id */
            name TEXT NOT NULL,     /* product name */
            price REAL              /* product price */
        );
        /* Another comment */
        """
        clean_sql = remove_sql_comments(sql_with_multi_comments)
        self.assertNotIn("/*", clean_sql)
        self.assertNotIn("*/", clean_sql)
        self.assertIn("CREATE TABLE", clean_sql)

        # Test mixed comments
        sql_with_mixed_comments = """
        -- Single line comment
        CREATE TABLE test (
            id INTEGER PRIMARY KEY, -- inline comment
            name TEXT NOT NULL      /* another comment */
        );
        /* Multi-line
           comment */
        SELECT * FROM test; -- end comment
        """
        clean_sql = remove_sql_comments(sql_with_mixed_comments)
        self.assertNotIn("--", clean_sql)
        self.assertNotIn("/*", clean_sql)
        self.assertNotIn("*/", clean_sql)
        self.assertIn("CREATE TABLE", clean_sql)
        self.assertIn("SELECT", clean_sql)

        # Test empty string
        self.assertEqual(remove_sql_comments(""), "")
        self.assertEqual(remove_sql_comments(None), "")

    def test_parse_sql_statements(self):
        """Test SQL statement parsing functionality."""
        # Test multiple statements
        multi_sql = """
        CREATE TABLE users (id INTEGER PRIMARY KEY, name TEXT);
        INSERT INTO users VALUES (1, 'John');
        INSERT INTO users VALUES (2, 'Jane');
        SELECT * FROM users;
        """
        statements = parse_sql_statements(multi_sql)
        self.assertEqual(len(statements), 4)
        self.assertIn("CREATE TABLE", statements[0])
        self.assertIn("INSERT INTO", statements[1])
        self.assertIn("INSERT INTO", statements[2])
        self.assertIn("SELECT", statements[3])

        # Test with semicolon stripping
        statements_no_semicolon = parse_sql_statements(multi_sql, strip_semicolon=True)
        self.assertEqual(len(statements_no_semicolon), 4)
        for stmt in statements_no_semicolon:
            self.assertFalse(stmt.endswith(";"))

        # Test with comments
        sql_with_comments = """
        -- Create table
        CREATE TABLE test (id INTEGER);
        /* Insert data */
        INSERT INTO test VALUES (1);
        SELECT * FROM test; -- Query data
        """
        statements = parse_sql_statements(sql_with_comments)
        self.assertEqual(len(statements), 3)
        self.assertIn("CREATE TABLE", statements[0])
        self.assertIn("INSERT INTO", statements[1])
        self.assertIn("SELECT", statements[2])

        # Test empty statements
        sql_with_empty = """
        CREATE TABLE test (id INTEGER);
        
        ;
        
        INSERT INTO test VALUES (1);
        """
        statements = parse_sql_statements(sql_with_empty)
        self.assertEqual(len(statements), 2)  # Empty statements filtered out

        # Test empty input
        self.assertEqual(parse_sql_statements(""), [])
        self.assertEqual(parse_sql_statements(None), [])

    def test_detect_statement_type_select(self):
        """Test statement type detection for SELECT statements."""
        # Basic SELECT
        self.assertEqual(detect_statement_type("SELECT * FROM users;"), "fetch")
        self.assertEqual(
            detect_statement_type("SELECT id, name FROM users WHERE id = 1;"), "fetch"
        )

        # SELECT with functions
        self.assertEqual(detect_statement_type("SELECT COUNT(*) FROM users;"), "fetch")
        self.assertEqual(
            detect_statement_type("SELECT AVG(salary) FROM employees;"), "fetch"
        )

        # SELECT with JOIN
        self.assertEqual(
            detect_statement_type(
                "SELECT u.name, p.title FROM users u JOIN posts p ON u.id = p.user_id;"
            ),
            "fetch",
        )

        # SELECT with subquery
        self.assertEqual(
            detect_statement_type(
                "SELECT * FROM users WHERE id IN (SELECT user_id FROM posts);"
            ),
            "fetch",
        )

    def test_detect_statement_type_cte(self):
        """Test statement type detection for CTEs (Common Table Expressions)."""
        # Simple CTE
        cte_sql = """
        WITH high_salary AS (
            SELECT * FROM employees WHERE salary > 80000
        )
        SELECT * FROM high_salary;
        """
        self.assertEqual(detect_statement_type(cte_sql), "fetch")

        # Multiple CTEs
        multi_cte_sql = """
        WITH dept_stats AS (
            SELECT department, COUNT(*) as count
            FROM employees 
            GROUP BY department
        ),
        high_count_depts AS (
            SELECT department
            FROM dept_stats 
            WHERE count > 5
        )
        SELECT * FROM high_count_depts;
        """
        self.assertEqual(detect_statement_type(multi_cte_sql), "fetch")

        # CTE with INSERT (should be execute)
        cte_insert_sql = """
        WITH new_emp AS (
            SELECT 'John' as name, 'Engineering' as dept
        )
        INSERT INTO employees (name, department)
        SELECT name, dept FROM new_emp;
        """
        self.assertEqual(detect_statement_type(cte_insert_sql), "execute")

    def test_detect_statement_type_cte_comprehensive(self):
        """Comprehensive CTE tests including WITH RECURSIVE and complex patterns."""

        # WITH RECURSIVE CTE
        recursive_cte_sql = """
        WITH RECURSIVE employee_hierarchy AS (
            SELECT id, name, manager_id, 1 as level
            FROM employees 
            WHERE manager_id IS NULL
            UNION ALL
            SELECT e.id, e.name, e.manager_id, eh.level + 1
            FROM employees e
            JOIN employee_hierarchy eh ON e.manager_id = eh.id
        )
        SELECT * FROM employee_hierarchy;
        """
        self.assertEqual(detect_statement_type(recursive_cte_sql), "fetch")

        # CTE with UPDATE
        cte_update_sql = """
        WITH high_salary_emps AS (
            SELECT id FROM employees WHERE salary > 100000
        )
        UPDATE employees 
        SET bonus = salary * 0.1 
        WHERE id IN (SELECT id FROM high_salary_emps);
        """
        self.assertEqual(detect_statement_type(cte_update_sql), "execute")

        # CTE with DELETE
        cte_delete_sql = """
        WITH inactive_users AS (
            SELECT id FROM users WHERE last_login < '2023-01-01'
        )
        DELETE FROM users WHERE id IN (SELECT id FROM inactive_users);
        """
        self.assertEqual(detect_statement_type(cte_delete_sql), "execute")

        # CTE with VALUES
        cte_values_sql = """
        WITH sample_data AS (
            VALUES 
                (1, 'Alice'),
                (2, 'Bob'),
                (3, 'Charlie')
        )
        SELECT * FROM sample_data;
        """
        self.assertEqual(detect_statement_type(cte_values_sql), "fetch")

        # Complex nested CTEs
        nested_cte_sql = """
        WITH 
        dept_summary AS (
            SELECT department, COUNT(*) as emp_count, AVG(salary) as avg_salary
            FROM employees 
            GROUP BY department
        ),
        high_avg_depts AS (
            SELECT department, avg_salary
            FROM dept_summary 
            WHERE avg_salary > 75000
        ),
        final_result AS (
            SELECT d.department, d.avg_salary, e.name as top_earner
            FROM high_avg_depts d
            JOIN employees e ON d.department = e.department
            WHERE e.salary = (
                SELECT MAX(salary) 
                FROM employees e2 
                WHERE e2.department = d.department
            )
        )
        SELECT * FROM final_result;
        """
        self.assertEqual(detect_statement_type(nested_cte_sql), "fetch")

        # CTE with window functions
        cte_window_sql = """
        WITH salary_ranks AS (
            SELECT 
                name,
                department,
                salary,
                ROW_NUMBER() OVER (PARTITION BY department ORDER BY salary DESC) as rank
            FROM employees
        )
        SELECT * FROM salary_ranks WHERE rank <= 3;
        """
        self.assertEqual(detect_statement_type(cte_window_sql), "fetch")

        # CTE with subqueries
        cte_subquery_sql = """
        WITH dept_leaders AS (
            SELECT 
                department,
                name,
                salary,
                (SELECT AVG(salary) FROM employees e2 WHERE e2.department = e1.department) as dept_avg
            FROM employees e1
            WHERE salary > (SELECT AVG(salary) FROM employees e3 WHERE e3.department = e1.department)
        )
        SELECT * FROM dept_leaders;
        """
        self.assertEqual(detect_statement_type(cte_subquery_sql), "fetch")

        # CTE with INSERT and multiple CTEs
        cte_multi_insert_sql = """
        WITH 
        new_employees AS (
            SELECT 'Alice' as name, 'Engineering' as dept, 85000 as salary
            UNION ALL
            SELECT 'Bob', 'Marketing', 65000
            UNION ALL
            SELECT 'Carol', 'Engineering', 90000
        ),
        valid_depts AS (
            SELECT DISTINCT department FROM employees WHERE department IS NOT NULL
        )
        INSERT INTO employees (name, department, salary)
        SELECT ne.name, ne.dept, ne.salary
        FROM new_employees ne
        JOIN valid_depts vd ON ne.dept = vd.department;
        """
        self.assertEqual(detect_statement_type(cte_multi_insert_sql), "execute")

        # CTE with UPDATE and complex conditions
        cte_complex_update_sql = """
        WITH 
        salary_stats AS (
            SELECT 
                department,
                AVG(salary) as avg_salary,
                STDDEV(salary) as salary_stddev
            FROM employees 
            GROUP BY department
        ),
        outliers AS (
            SELECT e.id, e.name, e.department, e.salary
            FROM employees e
            JOIN salary_stats ss ON e.department = ss.department
            WHERE e.salary > ss.avg_salary + (2 * ss.salary_stddev)
        )
        UPDATE employees 
        SET salary = salary * 0.9
        WHERE id IN (SELECT id FROM outliers);
        """
        self.assertEqual(detect_statement_type(cte_complex_update_sql), "execute")

        # CTE with DELETE and joins
        cte_delete_join_sql = """
        WITH 
        inactive_depts AS (
            SELECT department
            FROM employees e
            LEFT JOIN user_activity ua ON e.id = ua.employee_id
            GROUP BY department
            HAVING MAX(ua.last_activity) < '2023-01-01'
        )
        DELETE FROM employees 
        WHERE department IN (SELECT department FROM inactive_depts);
        """
        self.assertEqual(detect_statement_type(cte_delete_join_sql), "execute")

        # CTE with CASE statements
        cte_case_sql = """
        WITH salary_categories AS (
            SELECT 
                name,
                department,
                salary,
                CASE 
                    WHEN salary < 50000 THEN 'Low'
                    WHEN salary < 80000 THEN 'Medium'
                    ELSE 'High'
                END as category
            FROM employees
        )
        SELECT * FROM salary_categories WHERE category = 'High';
        """
        self.assertEqual(detect_statement_type(cte_case_sql), "fetch")

        # CTE with CTEs that have CTEs (nested CTE definitions)
        nested_cte_def_sql = """
        WITH 
        dept_data AS (
            WITH dept_summary AS (
                SELECT department, COUNT(*) as count
                FROM employees 
                GROUP BY department
            )
            SELECT department, count, 
                   CASE WHEN count > 10 THEN 'Large' ELSE 'Small' END as size
            FROM dept_summary
        )
        SELECT * FROM dept_data WHERE size = 'Large';
        """
        self.assertEqual(detect_statement_type(nested_cte_def_sql), "fetch")

    def test_detect_statement_type_dml(self):
        """Test statement type detection for DML statements."""
        # INSERT
        self.assertEqual(
            detect_statement_type("INSERT INTO users (name) VALUES ('John');"),
            "execute",
        )
        self.assertEqual(
            detect_statement_type("INSERT INTO users SELECT * FROM temp_users;"),
            "execute",
        )

        # UPDATE
        self.assertEqual(
            detect_statement_type("UPDATE users SET name = 'Jane' WHERE id = 1;"),
            "execute",
        )
        self.assertEqual(
            detect_statement_type("UPDATE users SET active = false;"), "execute"
        )

        # DELETE
        self.assertEqual(
            detect_statement_type("DELETE FROM users WHERE id = 1;"), "execute"
        )
        self.assertEqual(detect_statement_type("DELETE FROM users;"), "execute")

    def test_detect_statement_type_ddl(self):
        """Test statement type detection for DDL statements."""
        # CREATE
        self.assertEqual(
            detect_statement_type("CREATE TABLE users (id INTEGER PRIMARY KEY);"),
            "execute",
        )
        self.assertEqual(
            detect_statement_type("CREATE INDEX idx_name ON users (name);"), "execute"
        )
        self.assertEqual(
            detect_statement_type("CREATE VIEW user_view AS SELECT * FROM users;"),
            "execute",
        )

        # ALTER
        self.assertEqual(
            detect_statement_type("ALTER TABLE users ADD COLUMN email TEXT;"), "execute"
        )
        self.assertEqual(
            detect_statement_type("ALTER TABLE users DROP COLUMN email;"), "execute"
        )

        # DROP
        self.assertEqual(detect_statement_type("DROP TABLE users;"), "execute")
        self.assertEqual(detect_statement_type("DROP INDEX idx_name;"), "execute")

    def test_detect_statement_type_other(self):
        """Test statement type detection for other statement types."""
        # VALUES (some databases return rows)
        self.assertEqual(
            detect_statement_type("VALUES (1, 'John'), (2, 'Jane');"), "fetch"
        )

        # SHOW (information queries)
        self.assertEqual(detect_statement_type("SHOW TABLES;"), "fetch")
        self.assertEqual(detect_statement_type("SHOW COLUMNS FROM users;"), "fetch")

        # DESCRIBE/DESC
        self.assertEqual(detect_statement_type("DESCRIBE users;"), "fetch")
        self.assertEqual(detect_statement_type("DESC users;"), "fetch")

        # EXPLAIN
        self.assertEqual(detect_statement_type("EXPLAIN SELECT * FROM users;"), "fetch")

        # PRAGMA (SQLite metadata)
        self.assertEqual(detect_statement_type("PRAGMA table_info(users);"), "fetch")

        # BEGIN/COMMIT/ROLLBACK
        self.assertEqual(detect_statement_type("BEGIN TRANSACTION;"), "execute")
        self.assertEqual(detect_statement_type("COMMIT;"), "execute")
        self.assertEqual(detect_statement_type("ROLLBACK;"), "execute")

    def test_detect_statement_type_edge_cases(self):
        """Test statement type detection for edge cases."""
        # Empty or whitespace
        self.assertEqual(detect_statement_type(""), "execute")
        self.assertEqual(detect_statement_type("   "), "execute")
        self.assertEqual(detect_statement_type("\n\t"), "execute")

        # Case insensitive
        self.assertEqual(detect_statement_type("select * from users;"), "fetch")
        self.assertEqual(detect_statement_type("SELECT * FROM users;"), "fetch")
        self.assertEqual(detect_statement_type("Select * From users;"), "fetch")

        # With comments
        sql_with_comments = """
        -- This is a comment
        SELECT * FROM users; /* another comment */
        """
        self.assertEqual(detect_statement_type(sql_with_comments), "fetch")

        # Complex whitespace
        self.assertEqual(
            detect_statement_type("  SELECT  *  FROM  users  ;  "), "fetch"
        )

    def test_detect_statement_type_malformed_sql(self):
        """Test statement type detection with malformed or problematic SQL."""
        # Incomplete statements
        self.assertEqual(
            detect_statement_type("SELECT"), "fetch"
        )  # Incomplete but still SELECT
        self.assertEqual(detect_statement_type("WITH"), "execute")  # Incomplete WITH
        self.assertEqual(
            detect_statement_type("INSERT"), "execute"
        )  # Incomplete INSERT

        # Statements with only keywords
        self.assertEqual(detect_statement_type("SELECT SELECT"), "fetch")
        self.assertEqual(detect_statement_type("WITH WITH"), "execute")

        # Statements with special characters
        self.assertEqual(
            detect_statement_type("SELECT * FROM `users`;"), "fetch"
        )  # Backticks
        self.assertEqual(
            detect_statement_type('SELECT * FROM "users";'), "fetch"
        )  # Double quotes
        self.assertEqual(
            detect_statement_type("SELECT * FROM [users];"), "fetch"
        )  # Square brackets

        # Statements with numbers and symbols
        self.assertEqual(detect_statement_type("SELECT 1+1;"), "fetch")
        self.assertEqual(detect_statement_type("SELECT COUNT(*) FROM users;"), "fetch")

        # Very long statements (stress test)
        long_sql = "SELECT " + "a" * 1000 + " FROM users;"
        self.assertEqual(detect_statement_type(long_sql), "fetch")

    def test_detect_statement_type_complex_nested_structures(self):
        """Test statement type detection with very complex nested SQL structures."""
        # Deeply nested subqueries
        nested_subquery_sql = """
        SELECT * FROM users WHERE id IN (
            SELECT user_id FROM posts WHERE id IN (
                SELECT post_id FROM comments WHERE id IN (
                    SELECT comment_id FROM likes WHERE id IN (
                        SELECT like_id FROM reactions WHERE type = 'love'
                    )
                )
            )
        );
        """
        self.assertEqual(detect_statement_type(nested_subquery_sql), "fetch")

        # Complex CTE with multiple levels of nesting
        complex_nested_cte_sql = """
        WITH 
        level1 AS (
            SELECT id, name FROM users
        ),
        level2 AS (
            SELECT l1.id, l1.name, p.title 
            FROM level1 l1 
            JOIN posts p ON l1.id = p.user_id
        ),
        level3 AS (
            SELECT l2.id, l2.name, l2.title, c.content
            FROM level2 l2
            JOIN comments c ON l2.id = c.post_id
        ),
        level4 AS (
            SELECT l3.id, l3.name, l3.title, l3.content, l.count
            FROM level3 l3
            JOIN (
                SELECT post_id, COUNT(*) as count 
                FROM likes 
                GROUP BY post_id
            ) l ON l3.id = l.post_id
        )
        SELECT * FROM level4 WHERE count > 10;
        """
        self.assertEqual(detect_statement_type(complex_nested_cte_sql), "fetch")

        # CTE with UNION and complex joins
        cte_union_complex_sql = """
        WITH 
        active_users AS (
            SELECT id, name, 'active' as status FROM users WHERE last_login > '2023-01-01'
            UNION ALL
            SELECT id, name, 'recent' as status FROM users WHERE created_date > '2023-06-01'
        ),
        user_stats AS (
            SELECT 
                au.id,
                au.name,
                au.status,
                COUNT(p.id) as post_count,
                COUNT(c.id) as comment_count,
                AVG(p.rating) as avg_rating
            FROM active_users au
            LEFT JOIN posts p ON au.id = p.user_id
            LEFT JOIN comments c ON au.id = c.user_id
            GROUP BY au.id, au.name, au.status
        ),
        top_users AS (
            SELECT * FROM user_stats 
            WHERE post_count > 5 OR comment_count > 20 OR avg_rating > 4.0
        )
        SELECT 
            status,
            COUNT(*) as user_count,
            AVG(post_count) as avg_posts,
            AVG(comment_count) as avg_comments
        FROM top_users 
        GROUP BY status;
        """
        self.assertEqual(detect_statement_type(cte_union_complex_sql), "fetch")

    def test_detect_statement_type_database_specific_syntax(self):
        """Test statement type detection with database-specific SQL syntax."""
        # PostgreSQL specific
        postgres_sql = """
        WITH RECURSIVE employee_tree AS (
            SELECT id, name, manager_id, 1 as level
            FROM employees WHERE manager_id IS NULL
            UNION ALL
            SELECT e.id, e.name, e.manager_id, et.level + 1
            FROM employees e
            JOIN employee_tree et ON e.manager_id = et.id
        )
        SELECT * FROM employee_tree;
        """
        self.assertEqual(detect_statement_type(postgres_sql), "fetch")

        # SQLite specific
        sqlite_sql = """
        WITH RECURSIVE fibonacci(n, fib_n, fib_n_plus_1) AS (
            SELECT 0, 0, 1
            UNION ALL
            SELECT n + 1, fib_n_plus_1, fib_n + fib_n_plus_1
            FROM fibonacci WHERE n < 10
        )
        SELECT fib_n FROM fibonacci;
        """
        self.assertEqual(detect_statement_type(sqlite_sql), "fetch")

        # MySQL specific
        mysql_sql = """
        WITH RECURSIVE cte AS (
            SELECT 1 as n
            UNION ALL
            SELECT n + 1 FROM cte WHERE n < 5
        )
        SELECT * FROM cte;
        """
        self.assertEqual(detect_statement_type(mysql_sql), "fetch")

    def test_detect_statement_type_advanced_sql_features(self):
        """Test statement type detection with advanced SQL features."""
        # Window functions
        window_sql = """
        SELECT 
            name,
            department,
            salary,
            ROW_NUMBER() OVER (PARTITION BY department ORDER BY salary DESC) as rank,
            LAG(salary) OVER (PARTITION BY department ORDER BY hire_date) as prev_salary
        FROM employees;
        """
        self.assertEqual(detect_statement_type(window_sql), "fetch")

        # Pivot-like queries
        pivot_sql = """
        SELECT 
            department,
            SUM(CASE WHEN salary < 50000 THEN 1 ELSE 0 END) as low_salary,
            SUM(CASE WHEN salary BETWEEN 50000 AND 80000 THEN 1 ELSE 0 END) as mid_salary,
            SUM(CASE WHEN salary > 80000 THEN 1 ELSE 0 END) as high_salary
        FROM employees 
        GROUP BY department;
        """
        self.assertEqual(detect_statement_type(pivot_sql), "fetch")

        # JSON operations (modern SQL)
        json_sql = """
        SELECT 
            id,
            name,
            JSON_EXTRACT(metadata, '$.department') as dept,
            JSON_EXTRACT(metadata, '$.skills') as skills
        FROM users 
        WHERE JSON_CONTAINS(metadata, '"Python"', '$.skills');
        """
        self.assertEqual(detect_statement_type(json_sql), "fetch")

        # Full-text search
        fulltext_sql = """
        SELECT id, title, content, 
               MATCH(title, content) AGAINST('database optimization' IN NATURAL LANGUAGE MODE) as relevance
        FROM articles 
        WHERE MATCH(title, content) AGAINST('database optimization' IN NATURAL LANGUAGE MODE);
        """
        self.assertEqual(detect_statement_type(fulltext_sql), "fetch")

    def test_detect_statement_type_error_handling(self):
        """Test statement type detection with various error conditions."""
        # None input
        self.assertEqual(detect_statement_type(None), "execute")

        # Non-string input (should handle gracefully)
        try:
            result = detect_statement_type(123)
            # If it doesn't raise an exception, it should return 'execute'
            self.assertEqual(result, "execute")
        except (AttributeError, TypeError):
            # If it raises an exception, that's also acceptable behavior
            pass

        # Very large input
        large_sql = "SELECT " + "a" * 10000 + " FROM users;"
        self.assertEqual(detect_statement_type(large_sql), "fetch")

        # SQL with unusual token patterns
        unusual_sql = "SELECT * FROM users WHERE name LIKE '%test%' ESCAPE '\\';"
        self.assertEqual(detect_statement_type(unusual_sql), "fetch")

        # SQL with unicode characters
        unicode_sql = "SELECT * FROM users WHERE name = 'José';"
        self.assertEqual(detect_statement_type(unicode_sql), "fetch")

    def test_split_sql_file(self):
        """Test SQL file splitting functionality."""
        import tempfile
        import os

        # Create a temporary SQL file
        with tempfile.NamedTemporaryFile(mode="w", suffix=".sql", delete=False) as f:
            f.write(
                """
            -- Test SQL file
            CREATE TABLE test (id INTEGER PRIMARY KEY);
            INSERT INTO test VALUES (1);
            SELECT * FROM test;
            """
            )
            temp_file = f.name

        try:
            statements = split_sql_file(temp_file)
            self.assertEqual(len(statements), 3)
            self.assertIn("CREATE TABLE", statements[0])
            self.assertIn("INSERT INTO", statements[1])
            self.assertIn("SELECT", statements[2])

            # Test with semicolon stripping
            statements_no_semicolon = split_sql_file(temp_file, strip_semicolon=True)
            self.assertEqual(len(statements_no_semicolon), 3)
            for stmt in statements_no_semicolon:
                self.assertFalse(stmt.endswith(";"))

        finally:
            # Clean up
            os.unlink(temp_file)

        # Test file not found
        with self.assertRaises(FileNotFoundError):
            split_sql_file("nonexistent_file.sql")

        # Test invalid input
        with self.assertRaises(ValueError):
            split_sql_file("")
        with self.assertRaises(ValueError):
            split_sql_file(None)

    def test_detect_statement_type_cte_keyword_ambiguity_fix(self):
        """
        Test that CTE parsing correctly identifies statement types and doesn't
        incorrectly match non-statement keywords like FROM, WHERE, JOIN, etc.

        This test specifically validates the fix for the overly broad
        token.ttype in (DML, Keyword) check that could incorrectly identify
        keywords as statement types.
        """
        # CTE with complex structure that could trigger keyword confusion
        cte_with_complex_structure = """
        WITH user_data AS (
            SELECT id, name FROM users WHERE active = 1
        ),
        filtered_data AS (
            SELECT * FROM user_data WHERE name LIKE 'A%'
        )
        SELECT * FROM filtered_data WHERE id > 100;
        """
        # This should correctly identify as 'fetch' (SELECT), not incorrectly match 'FROM' or 'WHERE'
        self.assertEqual(detect_statement_type(cte_with_complex_structure), "fetch")

        # CTE with multiple JOINs and WHERE clauses that could be confused
        cte_with_joins_and_where = """
        WITH user_info AS (
            SELECT id, name FROM users WHERE active = 1
        ),
        post_info AS (
            SELECT user_id, title FROM posts WHERE published = 1
        )
        SELECT ui.name, pi.title 
        FROM user_info ui 
        JOIN post_info pi ON ui.id = pi.user_id 
        WHERE pi.title LIKE '%SQL%';
        """
        # This should correctly identify as 'fetch' (SELECT), not incorrectly match 'FROM', 'JOIN', or 'WHERE'
        self.assertEqual(detect_statement_type(cte_with_joins_and_where), "fetch")

        # CTE with GROUP BY, HAVING, ORDER BY that could be confused
        cte_with_aggregation_keywords = """
        WITH user_stats AS (
            SELECT 
                u.id, u.name, u.email,
                COUNT(p.id) as post_count,
                AVG(p.rating) as avg_rating
            FROM users u
            LEFT JOIN posts p ON u.id = p.user_id
            WHERE u.active = 1
            GROUP BY u.id, u.name, u.email
            HAVING COUNT(p.id) > 0
        )
        SELECT name, post_count, avg_rating
        FROM user_stats 
        WHERE avg_rating > 4.0 
        ORDER BY post_count DESC;
        """
        # This should correctly identify as 'fetch' (SELECT), not match any of the other keywords
        self.assertEqual(detect_statement_type(cte_with_aggregation_keywords), "fetch")

        # CTE with INSERT but complex structure that could trigger keyword confusion
        cte_insert_with_complex_structure = """
        WITH new_data AS (
            SELECT 'John' as name, 'Engineering' as dept
            UNION ALL
            SELECT 'Jane', 'Marketing'
        ),
        valid_depts AS (
            SELECT DISTINCT department FROM employees WHERE department IS NOT NULL
        )
        INSERT INTO employees (name, department) 
        SELECT nd.name, nd.dept
        FROM new_data nd
        JOIN valid_depts vd ON nd.dept = vd.department
        WHERE nd.dept = 'Engineering';
        """
        # This should correctly identify as 'execute' (INSERT), not match 'FROM', 'JOIN', or 'WHERE'
        self.assertEqual(
            detect_statement_type(cte_insert_with_complex_structure), "execute"
        )

        # CTE with UPDATE but complex structure that could trigger keyword confusion
        cte_update_with_complex_structure = """
        WITH target_employees AS (
            SELECT id FROM employees WHERE salary < 50000
        ),
        sales_employees AS (
            SELECT id FROM employees WHERE department = 'Sales'
        )
        UPDATE employees 
        SET salary = salary * 1.1
        WHERE id IN (
            SELECT te.id 
            FROM target_employees te
            JOIN sales_employees se ON te.id = se.id
        );
        """
        # This should correctly identify as 'execute' (UPDATE), not match 'FROM', 'JOIN', or 'WHERE'
        self.assertEqual(
            detect_statement_type(cte_update_with_complex_structure), "execute"
        )

        # CTE with complex nested structure and many keywords that could be confused
        cte_complex_nested_keywords = """
        WITH 
        base_data AS (
            SELECT id, name, department, salary FROM employees
        ),
        filtered_data AS (
            SELECT * FROM base_data WHERE salary > 50000
        ),
        aggregated_data AS (
            SELECT 
                department,
                COUNT(*) as count,
                AVG(salary) as avg_salary,
                MAX(salary) as max_salary
            FROM filtered_data 
            GROUP BY department
            HAVING COUNT(*) > 5
        )
        SELECT department, count, avg_salary
        FROM aggregated_data 
        WHERE avg_salary > 70000
        ORDER BY max_salary DESC;
        """
        # This should correctly identify as 'fetch' (SELECT), not match any of the other keywords
        self.assertEqual(detect_statement_type(cte_complex_nested_keywords), "fetch")

    def test_detect_statement_type_malformed_sql_edge_cases(self):
        # Test various malformed SQL that might exercise error paths
        self.assertEqual(detect_statement_type(""), "execute")
        self.assertEqual(detect_statement_type("   "), "execute")
        self.assertEqual(detect_statement_type("-- comment only"), "execute")
        self.assertEqual(detect_statement_type("/* comment only */"), "execute")

    def test_detect_statement_type_cte_fallback_scenarios(self):
        # Test CTE scenarios that trigger the fallback to _find_first_dml_keyword_top_level
        # This happens when _find_main_statement_after_ctes returns None
        cte_with_complex_structure = """
        WITH cte AS (
            SELECT 1 as x
        )
        SELECT * FROM cte;
        """
        self.assertEqual(detect_statement_type(cte_with_complex_structure), "fetch")

    def test_detect_statement_type_non_select_fetch_statements(self):
        # Test other fetch statement types that aren't SELECT
        self.assertEqual(detect_statement_type("VALUES (1, 2, 3);"), "fetch")
        self.assertEqual(detect_statement_type("SHOW TABLES;"), "fetch")
        self.assertEqual(detect_statement_type("EXPLAIN SELECT * FROM users;"), "fetch")
        self.assertEqual(detect_statement_type("PRAGMA table_info(users);"), "fetch")

    def test_detect_statement_type_modify_dml_types(self):
        # Test CTE with INSERT, UPDATE, DELETE to trigger the _MODIFY_DML_TYPES branch
        cte_insert = """
        WITH new_data AS (SELECT 1 as id)
        INSERT INTO table1 SELECT * FROM new_data;
        """
        self.assertEqual(detect_statement_type(cte_insert), "execute")

        cte_update = """
        WITH updates AS (SELECT 1 as id)
        UPDATE table1 SET col = 1 WHERE id IN (SELECT id FROM updates);
        """
        self.assertEqual(detect_statement_type(cte_update), "execute")

        cte_delete = """
        WITH to_delete AS (SELECT 1 as id)
        DELETE FROM table1 WHERE id IN (SELECT id FROM to_delete);
        """
        self.assertEqual(detect_statement_type(cte_delete), "execute")

    def test_detect_statement_type_other_fetch_statements_in_cte(self):
        # Test CTE with other fetch statements (not SELECT, INSERT, UPDATE, DELETE)
        # Note: VALUES after CTE might be parsed as execute depending on the parsing logic
        cte_values = """
        WITH cte AS (SELECT 1 as x)
        VALUES (1, 2, 3);
        """
        result = detect_statement_type(cte_values)
        # The actual behavior depends on how sqlparse handles VALUES after CTE
        # Let's accept either result as valid
        self.assertIn(result, ["fetch", "execute"])

    def test_parse_sql_statements_with_semicolon_stripping(self):
        # Test semicolon stripping functionality
        sql = "SELECT * FROM users; INSERT INTO users VALUES (1);"
        stmts = parse_sql_statements(sql, strip_semicolon=True)
        self.assertEqual(len(stmts), 2)
        self.assertFalse(stmts[0].endswith(";"))
        self.assertFalse(stmts[1].endswith(";"))

    def test_parse_sql_statements_empty_tokens(self):
        # Test case where tokens list is empty after flattening
        # This is hard to trigger with real SQL, but we can test the edge case
        sql = "   "  # Just whitespace
        stmts = parse_sql_statements(sql)
        self.assertEqual(stmts, [])

    def test_split_sql_file_os_error(self):
        # Test OSError handling in split_sql_file
        # This is hard to trigger reliably, but we can test the error path
        with self.assertRaises(ValueError):
            split_sql_file("")  # Empty string should raise ValueError

    def test_detect_statement_type_parsed_empty(self):
        # Test case where sqlparse.parse returns empty list
        # This is hard to trigger with normal SQL, but we can test the edge case
        # We'll use a very malformed SQL that sqlparse can't parse
        malformed_sql = "SELECT * FROM"  # Incomplete SQL
        # This should still return "execute" as it's not a complete statement
        result = detect_statement_type(malformed_sql)
        self.assertIn(result, ["fetch", "execute"])

    def test_detect_statement_type_tokens_empty_after_flatten(self):
        # Test case where stmt.flatten() returns empty tokens list
        # This is hard to trigger, but we can test the edge case
        empty_sql = "   "  # Just whitespace
        result = detect_statement_type(empty_sql)
        self.assertEqual(result, "execute")

    def test_detect_statement_type_first_token_none(self):
        # Test case where _next_non_ws_comment_token returns None for first_token
        # This happens when all tokens are whitespace or comments
        comment_only_sql = "-- comment\n/* block comment */"
        result = detect_statement_type(comment_only_sql)
        self.assertEqual(result, "execute")


# Additional tests for better coverage
class TestSqlHelperAdditional(unittest.TestCase):
    def test_split_sql_file_invalid_path(self):
        with self.assertRaises(FileNotFoundError):
            split_sql_file("nonexistent_file.sql")

    def test_split_sql_file_invalid_type(self):
        with self.assertRaises(ValueError):
            split_sql_file(None)
        with self.assertRaises(ValueError):
            split_sql_file(123)

    def test_parse_sql_statements_semicolons_in_strings(self):
        sql = """
        INSERT INTO test VALUES ('semicolon; inside string');
        SELECT * FROM test;
        """
        stmts = parse_sql_statements(sql)
        self.assertEqual(len(stmts), 2)
        self.assertIn("semicolon; inside string", stmts[0])

    def test_parse_sql_statements_only_comments(self):
        sql = """-- just a comment
        /* another comment */
        --;
        """
        stmts = parse_sql_statements(sql)
        self.assertEqual(stmts, [])

    def test_detect_statement_type_vendor_specific(self):
        self.assertEqual(detect_statement_type("SHOW TABLES;"), "fetch")
        self.assertEqual(detect_statement_type("PRAGMA table_info(users);"), "fetch")
        self.assertEqual(detect_statement_type("EXPLAIN SELECT * FROM users;"), "fetch")
        self.assertEqual(detect_statement_type("DESC users;"), "fetch")
        self.assertEqual(detect_statement_type("DESCRIBE users;"), "fetch")

    def test_parse_sql_statements_whitespace(self):
        self.assertEqual(parse_sql_statements("   "), [])
        self.assertEqual(parse_sql_statements("\n\n"), [])


class TestSqlHelperCoverage(unittest.TestCase):
    def test_parse_sql_statements_all_comments_semicolon(self):
        sql = '-- comment\n;\n/* block comment */\n;\n'
        stmts = parse_sql_statements(sql)
        self.assertEqual(stmts, [])

    def test_parse_sql_statements_empty(self):
        self.assertEqual(parse_sql_statements(None), [])
        self.assertEqual(parse_sql_statements(""), [])
        self.assertEqual(parse_sql_statements("   "), [])

    def test_parse_sql_statements_all_tokens_comments(self):
        sql = '-- comment\n/* block comment */\n'
        stmts = parse_sql_statements(sql)
        self.assertEqual(stmts, [])

    def test_detect_statement_type_complex_cte_edge_cases(self):
        # Test CTE with complex nested structure that exercises internal parsing
        complex_cte = """
        WITH RECURSIVE cte AS (
            SELECT 1 as n
            UNION ALL
            SELECT n + 1 FROM cte WHERE n < 10
        )
        SELECT * FROM cte;
        """
        self.assertEqual(detect_statement_type(complex_cte), "fetch")

    def test_detect_statement_type_cte_with_multiple_definitions(self):
        # Test CTE with multiple definitions separated by commas
        multi_cte = """
        WITH 
        cte1 AS (SELECT 1 as x),
        cte2 AS (SELECT 2 as y),
        cte3 AS (SELECT 3 as z)
        SELECT * FROM cte1, cte2, cte3;
        """
        self.assertEqual(detect_statement_type(multi_cte), "fetch")

    def test_detect_statement_type_cte_with_insert(self):
        # Test CTE that leads to an INSERT statement
        cte_insert = """
        WITH new_data AS (
            SELECT 'John' as name, 'Doe' as surname
        )
        INSERT INTO users (name, surname) 
        SELECT name, surname FROM new_data;
        """
        self.assertEqual(detect_statement_type(cte_insert), "execute")

    def test_detect_statement_type_malformed_sql_edge_cases(self):
        # Test various malformed SQL that might exercise error paths
        self.assertEqual(detect_statement_type(""), "execute")
        self.assertEqual(detect_statement_type("   "), "execute")
        self.assertEqual(detect_statement_type("-- comment only"), "execute")
        self.assertEqual(detect_statement_type("/* comment only */"), "execute")


if __name__ == "__main__":
    unittest.main()
